@@ -60,6 +60,10 @@
           (%with-array-data array index nil)
         (values vector index))
       (values (truly-the (simple-array * (*)) array) index)))
+
+(defun sb-c::%data-vector-and-index/check-bound (array index)
+  (%check-bound array (array-dimension array 0) index)
+  (%data-vector-and-index array index))
 
 
 ;;;; MAKE-ARRAY
@@ -702,12 +706,12 @@ of specialized arrays is supported."
                     (make-array ,(1+ widetag-mask)))
                 (declaim (type (simple-array function (,(1+ widetag-mask)))
                                ,table-name))
-                (defmacro ,name (array-var &optional vector-check)
-                  (if vector-check
+                (defmacro ,name (array-var &optional type)
+                  (if type
                       `(the function
                             (svref ,',table-name (%other-pointer-widetag
                                                   (locally (declare (optimize (safety 1)))
-                                                    (the vector ,array-var)))))
+                                                    (the ,type ,array-var)))))
                       `(the function
                             ;; Assigning TAG to 0 initially produces slightly better
                             ;; code than would be generated by the more natural expression
@@ -772,7 +776,11 @@ of specialized arrays is supported."
                 (defun ,(symbolicate 'vector- accessor-name) (array index ,@extra-params)
                   (declare (explicit-check)
                            (optimize speed (safety 0)))
-                  (%ref ,accessor-getter ,extra-params t))
+                  (%ref ,accessor-getter ,extra-params vector))
+                (defun ,(symbolicate 'string- accessor-name) (array index ,@extra-params)
+                  (declare (explicit-check)
+                           (optimize speed (safety 0)))
+                  (%ref ,accessor-getter ,extra-params string))
                 (defun ,slow-accessor-name (array index ,@extra-params)
                   (declare (optimize speed (safety 0))
                            (array array))
@@ -1160,26 +1168,6 @@ of specialized arrays is supported."
                     :format-arguments (list new max)))
            (setf (%array-fill-pointer vector) (truly-the index new))))))
 
-;;; FIXME: It'd probably make sense to use a MACROLET to share the
-;;; guts of VECTOR-PUSH between VECTOR-PUSH-EXTEND. Such a macro
-;;; should probably be based on the VECTOR-PUSH-EXTEND code (which is
-;;; new ca. sbcl-0.7.0) rather than the VECTOR-PUSH code (which dates
-;;; back to CMU CL).
-(defun vector-push (new-element array)
-  "Attempt to set the element of ARRAY designated by its fill pointer
-   to NEW-ELEMENT, and increment the fill pointer by one. If the fill pointer is
-   too large, NIL is returned, otherwise the index of the pushed element is
-   returned."
-  (declare (explicit-check))
-  (let ((fill-pointer (fill-pointer array)))
-    (cond ((= fill-pointer (%array-available-elements array))
-           nil)
-          (t
-           (locally (declare (optimize (safety 0)))
-             (setf (aref array fill-pointer) new-element))
-           (setf (%array-fill-pointer array) (1+ fill-pointer))
-           fill-pointer))))
-
 #-system-tlabs
 (defmacro reallocate-vector-with-widetag (old-vector &rest args)
   (declare (ignore old-vector))
@@ -1258,6 +1246,16 @@ of specialized arrays is supported."
       (setf (aref vector fill-pointer) new-element))
     fill-pointer))
 
+(defun prepare-vector-push-extend (vector)
+  (declare (explicit-check))
+  (let* ((fill-pointer (fill-pointer vector))
+         (new-fill-pointer (1+ fill-pointer)))
+    (if (= fill-pointer (%array-available-elements vector))
+        (extend-vector vector nil)
+        (setf (%array-fill-pointer vector) new-fill-pointer))
+    (multiple-value-bind (array index) (%data-vector-and-index vector fill-pointer)
+      (values array index fill-pointer))))
+
 (defun vector-pop (array)
   "Decrease the fill pointer by 1 and return the element pointed to by the
   new fill pointer."
@@ -1271,6 +1269,21 @@ of specialized arrays is supported."
                 (setf (%array-fill-pointer array)
                       (1- fill-pointer)))))))
 
+(defun vector-push (new-element array)
+  "Attempt to set the element of ARRAY designated by its fill pointer
+   to NEW-ELEMENT, and increment the fill pointer by one. If the fill pointer is
+   too large, NIL is returned, otherwise the index of the pushed element is
+   returned."
+  (declare (explicit-check))
+  (let ((fill-pointer (fill-pointer array)))
+    (cond ((= fill-pointer (%array-available-elements array))
+           nil)
+          (t
+           (locally (declare (optimize (safety 0)))
+             (setf (aref array fill-pointer) new-element))
+           (setf (%array-fill-pointer array) (1+ fill-pointer))
+           fill-pointer))))
+
 
 ;;;; ADJUST-ARRAY
 
