@@ -70,6 +70,8 @@
            function-with-layout-p
            non-null-symbol-p)
     (t) boolean (movable foldable flushable))
+(defknown unsigned-byte-x-p
+    (t (integer 1)) boolean (movable foldable flushable))
 
 (defknown car-eq-if-listp (t t) boolean (movable foldable flushable))
 
@@ -85,14 +87,15 @@
 (defknown %other-pointer-subtype-p (t list) boolean
   (movable foldable flushable always-translatable))
 
+(defknown string-designator-p (t) boolean
+  (movable foldable flushable always-translatable))
+
 ;;; Predicates that don't accept T for the first argument type
 (defknown (float-infinity-p float-nan-p float-infinity-or-nan-p)
   (float) boolean (movable foldable flushable))
 
 ;;;; miscellaneous "sub-primitives"
 
-(defknown pointer-hash (t) fixnum (flushable))
-
 (defknown %sp-string-compare
   (simple-string simple-string index (or null index) index (or null index))
   (values index fixnum)
@@ -102,36 +105,23 @@
   boolean
   (foldable flushable no-verify-arg-count))
 
-(defknown (sb-impl::instance-sxhash sb-impl::%instance-sxhash)
-    (instance) hash-code (flushable))
+(defknown sb-impl::instance-sxhash (instance) hash-code (flushable))
 ;;; SXHASH values on numbers and strings are predictable, therefore the next batch
 ;;; of functions are flushable. Perhaps not entirely obviously, symbol hashes are
-;;; predictable because we hash by name. And while ENSURE-SYMBOL-HASH is not
-;;; - strictly speaking - a side-effectless operation, it has no effect as far as
-;;; user-written code is concerned. i.e. nothing portably expressed can discern
-;;; whether memoization of the hash occurred. (I feel like we would have a cleaner
-;;; implementation if we just eagerly compute the hash anyway. Practically all
-;;; symbols eventually get hashed)
+;;; predictable because we hash by name.
 (defknown sb-impl::number-sxhash (number) hash-code (foldable flushable))
 (defknown %sxhash-string (string) hash-code (foldable flushable))
 (defknown %sxhash-simple-string (simple-string) hash-code (foldable flushable))
 
 (defknown (%sxhash-simple-substring) (simple-string index index) hash-code
   (foldable flushable))
-(defknown sb-impl::compute-symbol-hash (simple-string index) hash-code
-  (foldable flushable))
+(defknown sb-impl::calc-symbol-name-hash (simple-string index) symbol-name-hash ())
 
-(defknown (symbol-hash ensure-symbol-hash) (symbol) hash-code
-  (flushable movable))
-;;; This unusual accessor will read the word at SYMBOL-HASH-SLOT in any
-;;; object, not only symbols. The result is meaningful only if the object
-;;; is a symbol. The second argument indicates a predicate that the first
-;;; argument is known to satisfy, if any.
-;;; There is one more case, but an uninteresting one: the object is known
-;;; to be anything except NIL. That predicate would be IDENTITY,
-;;; which is not terribly helpful from a code generation stance.
-(defknown symbol-hash* (t (member nil symbolp non-null-symbol-p))
-  hash-code (flushable movable always-translatable))
+(defknown (symbol-hash) (symbol) hash-code (flushable movable))
+(defknown (symbol-name-hash) (symbol) symbol-name-hash (flushable movable))
+;;; This accessor will read the half-lispword at SYMBOL-HASH-SLOT in any object,
+;;; not only symbols. The value is reliable only if the object is a symbol.
+(defknown hash-as-if-symbol-name (t) symbol-name-hash (flushable movable always-translatable))
 
 (defknown %set-symbol-hash (symbol hash-code)
   t ())
@@ -157,7 +147,7 @@
 
 ;;; Return the length of VECTOR.
 ;;; Ordinary code should prefer to use (LENGTH (THE VECTOR FOO)) instead.
-(defknown vector-length (vector) index (flushable dx-safe))
+(defknown vector-length ((read-only vector)) index (flushable dx-safe))
 
 (defknown vector-sap ((simple-unboxed-array (*))) system-area-pointer
   (flushable))
@@ -184,13 +174,12 @@
 ;;; Like SET-HEADER-DATA, but instead of writing the entire header,
 ;;; LOGIOR of the specified value into the "data" portion of the word.
 ;;; Returns the first argument, *not* the modified header data.
-(defknown logior-header-bits (t (unsigned-byte 16)) t
+(defknown logior-header-bits (t (unsigned-byte 24)) (values)
     (#+x86-64 always-translatable))
 ;;; ASSIGN-VECTOR-FLAGSS assign all and only the flags byte.
-;;; RESET- performs LOGANDC2 and returns no value.
-(defknown (assign-vector-flags reset-header-bits)
-  (t (unsigned-byte 16)) (values)
-  (#+x86-64 always-translatable))
+(defknown assign-vector-flags (t (unsigned-byte 16)) (values) (#+x86-64 always-translatable))
+;;; RESET- performs LOGANDC2
+(defknown reset-header-bits (t (unsigned-byte 24)) (values) (#+x86-64 always-translatable))
 ;;; test bits of "HeaderData" which start 8 bits over from the lsb
 (defknown (test-header-data-bit)
   (t (unsigned-byte #.(- sb-vm:n-word-bits sb-vm:n-widetag-bits))) (boolean)
@@ -200,14 +189,16 @@
   (flushable))
 (defknown %set-array-dimension (array index index) (values)
   ())
-(defknown %array-rank (array) array-rank
+(defknown %array-rank (array) %array-rank
   (flushable))
 
-#+(or x86 x86-64 arm64)
 (defknown (%array-rank= widetag=) (t t) boolean
   (flushable))
 
-(defknown simple-array-header-of-rank-p (t array-rank) boolean
+(defknown vector-data (array index) (values simple-array index)
+  (flushable))
+
+(defknown simple-array-header-of-rank-p (t %array-rank) boolean
   (flushable))
 (defknown sb-kernel::check-array-shape (simple-array list)
   (simple-array)
@@ -688,15 +679,18 @@
 (defknown make-fdefn (t) fdefn (flushable movable))
 (defknown fdefn-p (t) boolean (movable foldable flushable))
 (defknown fdefn-name (fdefn) t (foldable flushable))
-(defknown fdefn-fun (fdefn) (or function null) (flushable))
+(defknown fdefn-fun ((or fdefn #+linkage-space symbol)) (or function null) (flushable))
 (defknown (setf fdefn-fun) (function fdefn) function ())
 (defknown fdefn-makunbound (fdefn) (values) ())
 ;;; FDEFN -> FUNCTION, trapping if not FBOUNDP
-(defknown safe-fdefn-fun (fdefn) function ())
+;;; For the most part this simple-fun is not needed, as ir2 conversion directly selects
+;;; the vop of this name; however, the expansion of handler-bind puts in a call to it
+;;; to trap unbound handlers in safe code, and we want that to work in the interpreter.
+(defknown safe-fdefn-fun ((or fdefn #+linkage-space symbol)) function ())
 
 (defknown %simple-fun-type (function) t (flushable))
 
-#+(or x86 x86-64 arm64) (defknown sb-vm::%closure-callee (function) fixnum (flushable))
+#+(or arm64 ppc64 x86 x86-64) (defknown sb-vm::%closure-callee (function) fixnum (flushable))
 (defknown %closure-fun (function) function (flushable))
 
 (defknown %closure-index-ref (function index) t
@@ -720,9 +714,29 @@
   sb-interpreter::interpreted-fun-prototype (flushable))
 
 
-(defknown %data-vector-and-index (array index)
-                                 (values (simple-array * (*)) index)
-                                 (foldable flushable))
+(defknown %data-vector-and-index
+    (array index)
+    (values (simple-array * (*)) index)
+    (foldable flushable))
+
+(defknown %data-vector-and-index/check-bound (array index)
+    (values (simple-array * (*)) index)
+    (foldable))
+
+(defknown %data-vector-and-index/known
+    (array index)
+    (values array index)
+    (flushable always-translatable))
+
+;;; Checks for and adjusts fill-pointer for vector-pop/push and
+;;; returns the underlying simple data vector.
+(defknown %data-vector-pop
+    (array)
+    (values (simple-array * (*)) index))
+
+(defknown %data-vector-push
+    (array)
+    (values (simple-array * (*)) (or null index)))
 
 (defknown restart-point (t) t ())
 
@@ -770,7 +784,7 @@
   (movable foldable flushable))
 
 (defknown (%sin %cos %tanh %sin-quick %cos-quick)
-  (double-float) (double-float $-1.0d0 $1.0d0)
+  (double-float) (double-float -1.0d0 1.0d0)
   (movable foldable flushable))
 
 (defknown (%asin %atan)
@@ -780,23 +794,23 @@
   (movable foldable flushable))
 
 (defknown (%acos)
-  (double-float) (double-float $0.0d0 #.(coerce pi 'double-float))
+  (double-float) (double-float 0.0d0 #.(coerce pi 'double-float))
   (movable foldable flushable))
 
 (defknown (%cosh)
-  (double-float) (double-float $1.0d0)
+  (double-float) (double-float 1.0d0)
   (movable foldable flushable))
 
 (defknown (%acosh %exp %sqrt)
-  (double-float) (double-float $0.0d0)
+  (double-float) (double-float 0.0d0)
   (movable foldable flushable))
 
 (defknown %expm1
-  (double-float) (double-float $-1d0)
+  (double-float) (double-float -1d0)
   (movable foldable flushable))
 
 (defknown (%hypot)
-  (double-float double-float) (double-float $0d0)
+  (double-float double-float) (double-float 0d0)
   (movable foldable flushable))
 
 (defknown (%pow)
@@ -817,7 +831,7 @@
   (double-float (signed-byte 32)) double-float
   (movable foldable flushable))
 
-(defknown (%log1p)
+(defknown (%log1p %log2)
   (double-float) double-float
   (movable foldable flushable))
 
