@@ -29,12 +29,12 @@
   (case (info :variable :kind name)
     (:special
      (let ((variable (find-free-var name)))
-       (make-lambda-var :%source-name name
+       (make-lambda-var name
                         :type (leaf-type variable)
                         :where-from (leaf-where-from variable)
                         :specvar variable)))
     (t
-     (make-lambda-var :%source-name name
+     (make-lambda-var name
                       :source-form source-form))))
 
 ;;; Parse a lambda list into a list of VAR structures, stripping off
@@ -63,7 +63,7 @@
                  (vars var)
                  var))
              (add-info (var kind &key (default nil defaultp) suppliedp-var key)
-               (let ((info (make-arg-info :kind kind)))
+               (let ((info (make-arg-info kind)))
                  (when defaultp
                    (setf (arg-info-default info) default
                          (arg-info-default-p info) t))
@@ -131,7 +131,8 @@
                                           :post-binding-lexenv post-binding-lexenv
                                           :debug-name (debug-name
                                                        '&aux-bindings
-                                                       aux-vars)
+                                                       (mapcar #'leaf-source-name
+                                                               aux-vars))
                                           :value-source-forms (rest value-source-forms))))
         (reference-leaf start ctran fun-lvar fun)
         (ir1-convert-combination-args fun-lvar ctran next result
@@ -162,7 +163,7 @@
                               :value-source-forms value-source-forms))
    (t
     (ctran-starts-block next)
-    (let ((cleanup (make-cleanup :kind :special-bind))
+    (let ((cleanup (make-cleanup :special-bind))
           (var (first svars))
           (bind-ctran (make-ctran))
           (cleanup-ctran (make-ctran)))
@@ -254,8 +255,8 @@
         (setf (node-lexenv bind) *lexenv*)
 
         (let ((block (ctran-starts-block result-ctran)))
-          (let ((return (make-return :result result-lvar :lambda lambda))
-                (tail-set (make-tail-set :funs (list lambda))))
+          (let ((return (make-return result-lvar lambda))
+                (tail-set (make-tail-set (list lambda))))
             (setf (lambda-tail-set lambda) tail-set)
             (setf (lambda-return lambda) return)
             (setf (lvar-dest result-lvar) return)
@@ -285,7 +286,7 @@
 (defun register-entry-point (entry dispatcher)
   (declare (type clambda entry)
            (type optional-dispatch dispatcher))
-  (setf (functional-kind entry) :optional)
+  (setf (functional-kind entry) (functional-kind-attributes optional))
   (setf (leaf-ever-used entry) t)
   (setf (lambda-optional-dispatch entry) dispatcher)
   entry)
@@ -304,7 +305,7 @@
   (let* ((fvars (reverse vars))
          (arg-vars (mapcar (lambda (var)
                              (make-lambda-var
-                              :%source-name (leaf-source-name var)
+                              (leaf-source-name var)
                               :type (leaf-type var)
                               :where-from (leaf-where-from var)
                               :specvar (lambda-var-specvar var)))
@@ -392,7 +393,7 @@
     (let ((name (or debug-name source-name)))
       (if (or force
               supplied-p-p ; this entry will be of kind NIL
-              (and (lambda-p ep) (eq (lambda-kind ep) nil)))
+              (and (lambda-p ep) (functional-kind-eq ep nil)))
           (convert-optional-entry ep
                                   default-vars default-vals
                                   (if supplied-p (list default nil) (list default))
@@ -448,18 +449,18 @@
             (body))
 
     (dolist (var (reverse entry-vars))
-      (arg-vars (make-lambda-var :%source-name (leaf-source-name var)
+      (arg-vars (make-lambda-var (leaf-source-name var)
                                  :type (leaf-type var)
                                  :where-from (leaf-where-from var))))
 
     (let* ((*allow-instrumenting* nil)
            (n-context (gensym "N-CONTEXT-"))
-           (context-temp (make-lambda-var :%source-name n-context
-                                          :arg-info (make-arg-info :kind :more-context)))
+           (context-temp (make-lambda-var n-context
+                                          :arg-info (make-arg-info :more-context)))
            (n-count (gensym "N-COUNT-"))
-           (count-temp (make-lambda-var :%source-name n-count
+           (count-temp (make-lambda-var n-count
                                         :type (specifier-type 'index)
-                                        :arg-info (make-arg-info :kind :more-count))))
+                                        :arg-info (make-arg-info :more-count))))
 
       (arg-vars context-temp count-temp)
 
@@ -486,16 +487,12 @@
               (found-allow-p nil))
 
           (temps #-stack-grows-downward-not-upward
-                 `(,n-index (+ ,n-count ,(if (vop-existsp :translate %more-keyword-pair)
+                 `(,n-index (+ ,n-count ,(if (vop-existsp :translate %more-kw-arg)
                                              0
                                              -1)))
                  #+stack-grows-downward-not-upward
-                 `(,n-index (- (1- ,n-count)))
-                 #-stack-grows-downward-not-upward n-value-temp
-                 #-stack-grows-downward-not-upward n-key)
-          (body `(declare (fixnum ,n-index)
-                          #-stack-grows-downward-not-upward
-                          (ignorable ,n-value-temp ,n-key)))
+                 `(,n-index (- (1- ,n-count))))
+          (body `(declare (fixnum ,n-index)))
 
           (collect ((tests))
             (dolist (key keys)
@@ -544,9 +541,11 @@
                        (setq ,n-lose ,n-key))))
 
             (body
-             `(when (oddp ,(if (vop-existsp :translate %more-keyword-pair)
-                               n-index
-                               n-count))
+             `(when (oddp ,(cond #-stack-grows-downward-not-upward
+                                 ((vop-existsp :translate %more-kw-arg)
+                                  n-index)
+                                 (t
+                                  n-count)))
                 (%odd-key-args-error)))
 
             (body
@@ -554,19 +553,20 @@
              `(locally
                 (declare (optimize (safety 0)))
                 (loop
-                 ,@(cond ((vop-existsp :translate %more-keyword-pair)
-                          `((when (zerop ,n-index) (return))
-                            (decf ,n-index 2)
-                            (multiple-value-bind (key value)
-                                (%more-keyword-pair ,n-context ,n-index)
-                              (setf ,n-value-temp value ,n-key key))))
-                         (t
-                          `((when (minusp ,n-index) (return))
-                            (setf ,n-value-temp (%more-arg ,n-context ,n-index))
-                            (decf ,n-index)
-                            (setq ,n-key (%more-arg ,n-context ,n-index))
-                            (decf ,n-index))))
-                 (case ,n-key ,@(tests))))
+                  ,@(cond ((vop-existsp :translate %more-kw-arg)
+                           `((when (zerop ,n-index) (return))
+                             (decf ,n-index 2)
+                             (multiple-value-bind (,n-value-temp ,n-key)
+                                 (%more-kw-arg ,n-context ,n-index)
+                               (declare (ignorable ,n-value-temp ,n-key))
+                               (case ,n-key ,@(tests)))))
+                          (t
+                           `((when (minusp ,n-index) (return))
+                             (let ((,n-value-temp (%more-arg ,n-context ,n-index))
+                                   (,n-key (%more-arg ,n-context (decf ,n-index))))
+                               (declare (ignorable ,n-value-temp ,n-key))
+                               (decf ,n-index)
+                               (case ,n-key ,@(tests))))))))
              #+stack-grows-downward-not-upward
              `(locally (declare (optimize (safety 0)))
                 (loop
@@ -633,11 +633,11 @@
         ;; ARG-INFO-DEFAULT for transforming (VALUES-LIST REST) into
         ;; (%MORE-ARG-VALUES CONTEXT 0 COUNT) when possible.
         (let* ((context-name (gensym "REST-CONTEXT-"))
-               (context (make-lambda-var :%source-name context-name
-                                         :arg-info (make-arg-info :kind :more-context)))
+               (context (make-lambda-var context-name
+                                         :arg-info (make-arg-info :more-context)))
                (count-name (gensym "REST-COUNT-"))
-               (count (make-lambda-var :%source-name count-name
-                                       :arg-info (make-arg-info :kind :more-count)
+               (count (make-lambda-var count-name
+                                       :arg-info (make-arg-info :more-count)
                                        :type (specifier-type 'index))))
           (setf (arg-info-default (lambda-var-arg-info rest)) (list context count)
                 (lambda-var-ever-used context) t
@@ -657,15 +657,14 @@
              (supplied-p (arg-info-supplied-p info))
              ;; was: (format nil "~A-DEFAULTING-TEMP" (leaf-source-name key))
              (n-val (make-symbol ".DEFAULTING-TEMP."))
-             (val-temp (make-lambda-var :%source-name n-val))
+             (val-temp (make-lambda-var n-val))
              (default `(with-source-form ,(lambda-var-source-form key)
                          ,default)))
         (main-vars val-temp)
         (bind-vars key)
         (cond ((or hairy-default supplied-p)
                (let* ((n-supplied (gensym "N-SUPPLIED-"))
-                      (supplied-temp (make-lambda-var
-                                      :%source-name n-supplied)))
+                      (supplied-temp (make-lambda-var n-supplied)))
                  (unless supplied-p
                    (setf (arg-info-supplied-p info) supplied-temp))
                  (when hairy-default
@@ -849,12 +848,12 @@
                                  debug-name)
   (declare (list body vars aux-vars aux-vals))
   (aver (or debug-name (neq '.anonymous. source-name)))
-  (let ((res (make-optional-dispatch :arglist vars
-                                     :allowp allowp
-                                     :keyp keyp
-                                     :%source-name source-name
-                                     :%debug-name debug-name
-                                     :source-path *current-path*))
+  (let ((res (make-optional-dispatch vars
+                                     allowp
+                                     keyp
+                                     source-name
+                                     debug-name
+                                     *current-path*))
         (min (or (position-if #'lambda-var-arg-info vars) (length vars))))
     (aver-live-component *current-component*)
     (ir1-convert-hairy-args res () () () () vars nil body aux-vars aux-vals
@@ -873,11 +872,16 @@
 
 (defun add-types-for-fixed-args (fun vars)
   (let ((fun-info (info :function :info fun)))
-    (when (and fun-info
-               (ir1-attributep (fun-info-attributes fun-info) fixed-args))
-      (loop for type in (fun-type-required (info :function :type fun))
+    (when (or (and fun-info
+                   (ir1-attributep (fun-info-attributes fun-info) fixed-args))
+              (typep fun '(cons (eql sb-impl::specialized-xep))))
+      (loop for type in (fun-type-required (if (typep fun '(cons (eql sb-impl::specialized-xep)))
+                                               (specifier-type `(function ,@(cddr fun)))
+                                               (info :function :type fun)))
             for var in vars
-            do (setf (lambda-var-type var) type))))
+            for intersection = (type-intersection type (lambda-var-type var))
+            unless (eq intersection *empty-type*)
+            do (setf (lambda-var-type var) intersection))))
   vars)
 
 ;;; Convert a LAMBDA form into a LAMBDA leaf or an OPTIONAL-DISPATCH leaf.
@@ -982,14 +986,6 @@
              (progn
                ,@forms))))))))
 
-;; FIXME: really should be an aspect of the lexical environment,
-;; but LEXENVs don't know whether they are toplevel or not.
-(defun has-toplevelness-decl (lambda-expr)
-  (dolist (expr (cddr lambda-expr)) ; Skip over (LAMBDA (ARGS))
-    (cond ((equal expr '(declare (top-level-form))) (return t))
-          ((typep expr '(or (cons (eql declare)) string))) ; DECL | DOCSTRING
-          (t (return nil)))))
-
 ;;; helper for LAMBDA-like things, to massage them into a form
 ;;; suitable for IR1-CONVERT-LAMBDA.
 (defun ir1-convert-lambdalike (thing
@@ -1007,64 +1003,37 @@
     ((named-lambda)
      (let* ((name (cadr thing))
             (lambda-expression `(lambda ,@(cddr thing)))
-            (*inline-expansions* (list name 1 *inline-expansions*))
-            (simple-lexenv-p (simple-lexical-environment-p *lexenv*)))
-       ;; Discard any forward references to this function unless we
-       ;; are block compiling and the lexical environment doesn't
-       ;; contain any hair. If the lexical environment is too hairy, then we only
-       ;; install the definition during the processing of this NAMED-LAMBDA,
-       ;; ensuring that the function cannot be called outside of the correct
-       ;; environment.
-       (unless (and (block-compile *compilation*) simple-lexenv-p)
-         (remhash name (free-funs *ir1-namespace*)))
+            (*inline-expansions* (list name 1 *inline-expansions*)))
        (if (and name (legal-fun-name-p name))
-           (let ((defined-fun-res (get-defined-fun name (second lambda-expression)))
-                 (res (ir1-convert-lambda lambda-expression
-                                          :maybe-add-debug-catch t
-                                          :source-name name))
-                 (info (info :function :info name)))
-             (setf (functional-inlinep res) (info :function :inlinep name)
-                   (defined-fun-same-block-p defined-fun-res) t)
-             (when (has-toplevelness-decl lambda-expression)
-               (setf (functional-top-level-defun-p res) t))
-             ;; FIXME: Should non-entry block compiled defuns have
-             ;; this propagate?
-             (unless (and info
-                          (ir1-attributep (fun-info-attributes info) fixed-args))
-               (assert-global-function-definition-type name res))
-             ;; If in a simple environment, then we can allow
-             ;; backward references to this function from following
-             ;; top-level forms.
-             (when simple-lexenv-p
-               (setf (defined-fun-functional defined-fun-res) res))
-             (unless (or
-                      (eq (defined-fun-inlinep defined-fun-res) 'notinline)
-                      ;; Don't treat recursive stubs like CAR as self-calls
-                      ;; Maybe just use the fact that it is a known function?
-                      ;; Though a known function may be used
-                      ;; because of some other attributues but
-                      ;; still wants to get optimized self calls
-                      (and info
-                           (or (fun-info-templates info)
-                               (fun-info-transforms info)
-                               (fun-info-ltn-annotate info)
-                               (fun-info-ir2-convert info)
-                               (fun-info-optimizer info))))
-               (if (block-compile *compilation*)
-                   (substitute-leaf res defined-fun-res)
-                   (substitute-leaf-if
-                    (lambda (ref)
-                      (policy ref (> recognize-self-calls 0)))
-                    res defined-fun-res)))
-             res)
+           (let ((simple-lexenv-p (simple-lexical-environment-p *lexenv*)))
+             ;; If not in a simple environment, then discard any
+             ;; forward references to this function. If the lexical
+             ;; environment is too hairy, then we only install the
+             ;; definition during the processing of this NAMED-LAMBDA,
+             ;; ensuring that the function cannot be called outside of
+             ;; the correct environment. If the function is globally
+             ;; NOTINLINE, then that inhibits even local substitution.
+             (unless simple-lexenv-p
+               (remhash name (free-funs *ir1-namespace*)))
+             (let ((var (get-defined-fun name)))
+               (setf (defined-fun-same-block-p var) t)
+               ;;
+               ;; If there is a type from a previous definition, blast it, since it is
+               ;; obsolete.
+               (when (neq :declared (leaf-where-from var))
+                 (setf (leaf-type var)
+                       ;; Use the type from the lambda list so that self
+                       ;; calls warn about mismatched args.
+                       (let ((lambda-list (second lambda-expression)))
+                         (or (and lambda-list
+                                  (ignore-errors
+                                   (ftype-from-lambda-list lambda-list)))
+                             (specifier-type 'function)))))
+               (ir1-convert-lambda-for-defun lambda-expression var simple-lexenv-p)))
            (ir1-convert-lambda lambda-expression
                                :maybe-add-debug-catch t
                                :debug-name
-                               (or name (name-lambdalike thing))))))
-    ((lambda-with-lexenv)
-     (ir1-convert-inline-lambda thing
-                                :source-name source-name
-                                :debug-name debug-name))))
+                               (or name (name-lambdalike thing))))))))
 
 (declaim (end-block))
 
@@ -1142,48 +1111,14 @@
             (t
              (let ((expansion (sb-walker:macroexpand-all lambda lexenv)))
                (if decls
-                   `(lambda-with-lexenv (:declare ,decls) ,@(cdr expansion))
+                   `(lambda-with-lexenv ((declare ,@decls)) ,@(cdr expansion))
                    expansion))))))
    #+(and sb-fasteval (not sb-xc-host))
    (sb-interpreter:basic-env
-    (awhen (sb-interpreter::reconstruct-syntactic-closure-env lexenv)
-      `(lambda-with-lexenv ,it ,@(cdr lambda))))
+    (sb-interpreter::inline-syntactic-closure-lambda lambda lexenv))
    #+sb-fasteval
    (null lambda))) ; trivial case. Never occurs in the compiler.
 
-(declaim (start-block ir1-convert-inline-lambda))
-
-;;; Convert the forms produced by RECONSTRUCT-LEXENV to LEXENV
-(defun process-inline-lexenv (inline-lexenv)
-  (labels ((recurse (inline-lexenv lexenv)
-             (let ((*lexenv* lexenv))
-               (if (null inline-lexenv)
-                   lexenv
-                   (destructuring-bind (type bindings &optional body) inline-lexenv
-                     (case type
-                       (:declare
-                        (recurse body
-                                 (process-decls `((declare ,@bindings)) nil nil)))
-                       (:macro
-                        (let ((macros
-                               (mapcar (lambda (binding)
-                                         ;; XC compile-in-lexenv ignores its second arg
-                                         #+sb-xc-host (aver (null-lexenv-p lexenv))
-                                         (list* (car binding) 'macro
-                                                (compile-in-lexenv (cdr binding) lexenv
-                                                                   nil nil nil t nil)))
-                                       bindings)))
-                          (recurse body
-                                   (make-lexenv :default lexenv
-                                                :funs macros))))
-                       (:symbol-macro
-                        (funcall-in-symbol-macrolet-lexenv bindings
-                                                           (lambda (&optional vars)
-                                                             (declare (ignore vars))
-                                                             (recurse body *lexenv*))
-                                                           :compile))))))))
-    (recurse inline-lexenv (make-null-lexenv))))
-
 ;;; Convert FUN as a lambda in the null environment, but use the
 ;;; current compilation policy. Note that FUN may be a
 ;;; LAMBDA-WITH-LEXENV, so we may have to augment the environment to
@@ -1191,46 +1126,52 @@
 (defun ir1-convert-inline-lambda (fun
                                   &key
                                   (source-name '.anonymous.)
-                                  debug-name
-                                  (policy (lexenv-policy *lexenv*)))
+                                  debug-name)
   (when (and (not debug-name) (eq '.anonymous. source-name))
     (setf debug-name (name-lambdalike fun)))
-  (let* ((lambda-with-lexenv-p (eq (car fun) 'lambda-with-lexenv))
-         (body (if lambda-with-lexenv-p
-                   `(lambda ,@(cddr fun))
-                   fun))
-         (lexenv-lambda (lexenv-lambda *lexenv*))
-         (*lexenv*
-           (if lambda-with-lexenv-p
-               (make-lexenv
-                :default (process-inline-lexenv (second fun))
-                :handled-conditions (lexenv-handled-conditions *lexenv*)
-                :policy policy
-                :flushable (lexenv-flushable *lexenv*)
-                :lambda lexenv-lambda
-                :parent *lexenv*)
-               (make-almost-null-lexenv
-                policy
-                ;; Inherit MUFFLE-CONDITIONS from the call-site lexenv
-                ;; rather than the definition-site lexenv, since it seems
-                ;; like a much more common case.
-                (lexenv-handled-conditions *lexenv*)
-                (lexenv-flushable *lexenv*)
-                lexenv-lambda
-                *lexenv*)))
-         (*inlining* (1+ *inlining*))
-         (clambda (ir1-convert-lambda body
-                                      :source-name source-name
-                                      :debug-name debug-name)))
-    (setf (functional-inline-expanded clambda) t)
-    clambda))
-
-(declaim (end-block))
+  (destructuring-bind (decls &rest body)
+      (if (eq (car fun) 'lambda-with-lexenv)
+          (cdr fun)
+          `(() . ,(cdr fun)))
+    (let* ((notinlines
+             (loop for fun in (lexenv-funs *lexenv*)
+                   when (and (defined-fun-p (cdr fun))
+                             (defined-fun-inlinep (cdr fun)))
+                   collect fun))
+           (*lexenv*
+             (if decls
+                 (make-lexenv
+                  :default (process-decls decls nil nil
+                                          :lexenv (make-null-lexenv))
+                  ;; Inherit MUFFLE-CONDITIONS from the call-site lexenv
+                  ;; rather than the definition-site lexenv, since it seems
+                  ;; like a much more common case.
+                  :handled-conditions (lexenv-handled-conditions *lexenv*)
+                  :policy (lexenv-policy *lexenv*)
+                  :flushable (lexenv-flushable *lexenv*)
+                  :lambda (lexenv-lambda *lexenv*)
+                  :parent *lexenv*)
+                 (make-almost-null-lexenv
+                  (lexenv-policy *lexenv*)
+                  (lexenv-handled-conditions *lexenv*)
+                  (lexenv-flushable *lexenv*)
+                  (lexenv-lambda *lexenv*)
+                  *lexenv*)))
+           (*inlining* (1+ *inlining*))
+           (clambda (progn
+                      (when notinlines
+                        (setf (lexenv-funs *lexenv*)
+                              notinlines))
+                      (ir1-convert-lambda `(lambda ,@body)
+                                          :source-name source-name
+                                          :debug-name debug-name))))
+      (setf (functional-inline-expanded clambda) t)
+      clambda)))
 
 ;;; Get a DEFINED-FUN object for a function we are about to define. If
 ;;; the function has been forward referenced, then substitute for the
 ;;; previous references.
-(defun get-defined-fun (name &optional (lambda-list nil lp))
+(defun get-defined-fun (name)
   (proclaim-as-fun-name name)
   (let ((found (find-free-fun name "shouldn't happen! (defined-fun)"))
         (free-funs (free-funs *ir1-namespace*)))
@@ -1246,23 +1187,20 @@
                      (not (info :function :inlinep name))))
            (let* ((where-from (leaf-where-from found))
                   (res (make-defined-fun
-                        :%source-name name
-                        :where-from (if (eq where-from :declared)
-                                        :declared
-                                        :defined-here)
-                        :type (if (eq :declared where-from)
-                                  (leaf-type found)
-                                  (or (and lp
-                                           (ignore-errors
-                                            (ftype-from-lambda-list lambda-list)))
-                                      (specifier-type 'function))))))
+                        name
+                        (if (eq where-from :declared-verify)
+                            (leaf-defined-type found)
+                            (leaf-type found))
+                        (if (memq where-from '(:declared :declared-verify))
+                            :declared
+                            :defined-here))))
              (substitute-leaf res found)
              (setf (gethash name free-funs) res)))
           ;; If FREE-FUNS has a previously converted definition
           ;; for this name, then blow it away and try again.
           ((defined-fun-functional found)
            (remhash name free-funs)
-           (get-defined-fun name lambda-list))
+           (get-defined-fun name))
           (t found))))
 
 ;;; Check a new global function definition for consistency with
@@ -1271,10 +1209,8 @@
 ;;; EXPLICIT-CHECK attribute, which is specified on functions that
 ;;; check their argument types as a consequence of type dispatching.
 ;;; This avoids redundant checks such as NUMBERP on the args to +, etc.
-;;; FIXME: this seems to have nothing at all to do with adding "new"
-;;; definitions, as it is only called from IR1-CONVERT-INLINE-EXPANSION.
 (defun assert-new-definition (var fun)
-  (let* ((type (leaf-type var))
+  (let* ((type (massage-global-definition-type (leaf-type var) fun))
          (for-real (eq (leaf-where-from var) :declared))
          (name (leaf-source-name var))
          (info (info :function :info name))
@@ -1310,56 +1246,94 @@
                 (consp (cdr entry)))
               (lexenv-vars lexenv))))
 
-;;; Used for global inline expansion. Earlier something like this was
-;;; used by %DEFUN too. FIXME: And now it's probably worth rethinking
-;;; whether this function is a good idea at all.
-(defun ir1-convert-inline-expansion (name expansion var inlinep info)
-  ;; Unless a INLINE function, we temporarily clobber the inline
-  ;; expansion. This prevents recursive inline expansion of
-  ;; opportunistic pseudo-inlines.
-  (unless (eq inlinep 'inline)
-    (setf (defined-fun-inline-expansion var) nil))
-  (let ((fun (ir1-convert-inline-lambda expansion
-                                        :source-name name)))
-    (setf (functional-inlinep fun) inlinep)
-    (assert-new-definition var fun)
-    (setf (defined-fun-inline-expansion var) expansion)
-    ;; Associate VAR with the FUN -- and in case of an optional dispatch
-    ;; with the various entry-points. This allows XREF to know where the
-    ;; inline CLAMBDA comes from.
-    (flet ((note-inlining (f)
-             (typecase f
-               (functional
-                (setf (functional-inline-expanded f) var))
-               (cons
-                ;; Delayed entry-point.
-                (if (car f)
-                    (setf (functional-inline-expanded (cdr f)) var)
-                    (let ((old-thunk (cdr f)))
-                      (setf (cdr f) (lambda ()
-                                      (let ((g (funcall old-thunk)))
-                                        (setf (functional-inline-expanded g) var)
-                                        g)))))))))
-      (note-inlining fun)
-      (when (optional-dispatch-p fun)
-        (note-inlining (optional-dispatch-main-entry fun))
-        (note-inlining (optional-dispatch-more-entry fun))
-        (mapc #'note-inlining (optional-dispatch-entry-points fun))))
-    ;; substitute for any old references
+;; FIXME: really should be an aspect of the lexical environment,
+;; but LEXENVs don't know whether they are toplevel or not.
+(defun has-toplevelness-decl (lambda-expr)
+  (dolist (expr (cddr lambda-expr)) ; Skip over (LAMBDA (ARGS))
+    (cond ((equal expr '(declare (top-level-form))) (return t))
+          ((typep expr '(or (cons (eql declare)) string))) ; DECL | DOCSTRING
+          (t (return nil)))))
+
+;;; Convert a lambda doing all the basic stuff we would do if we were
+;;; converting a DEFUN.
+(defun ir1-convert-lambda-for-defun (lambda var simple-lexenv-p)
+  (let* ((name (leaf-source-name var))
+         (fun (ir1-convert-lambda lambda
+                                  :maybe-add-debug-catch t
+                                  :source-name name))
+         (info (info :function :info name)))
+    (setf (functional-inlinep fun) (info :function :inlinep name))
+    (unless (and info
+                 (ir1-attributep (fun-info-attributes info) fixed-args))
+      (assert-new-definition var fun))
+    (when (has-toplevelness-decl lambda)
+      (setf (functional-top-level-defun-p fun) t))
+    ;; If definitely not an interpreter stub, then substitute for any
+    ;; old references.
     (unless (or (eq (defined-fun-inlinep var) 'notinline)
-                (not (block-compile *compilation*))
                 (and info
                      (or (fun-info-transforms info)
                          (fun-info-templates info)
                          (fun-info-ir2-convert info))))
-      (substitute-leaf fun var)
-      ;; If in a simple environment, then we can allow backward
-      ;; references to this function from following top-level
-      ;; forms.
-      (when (simple-lexical-environment-p *lexenv*)
-        (setf (defined-fun-functional var) fun)))
+      (let (type)
+        (if (block-compile *compilation*)
+            (progn
+              (substitute-leaf fun var)
+              ;; If in a simple environment, then we can allow backward
+              ;; references to this function from following top-level
+              ;; forms.
+              (when simple-lexenv-p
+                (setf (defined-fun-functional var) fun)))
+            (substitute-leaf-if
+             (lambda (ref)
+               (if (policy ref (> recognize-self-calls 0))
+                   t
+                   (let ((type (or type
+                                   (setf type (definition-type fun))))
+                         (call (node-dest ref)))
+                     (when (and (combination-p call)
+                                (eq (combination-fun call)
+                                    (ref-lvar ref))
+                                (fun-type-p type))
+                       (assert-call-type call type t :defined-here))
+                     nil)))
+             fun var))))
     fun))
 
+;;; Convert a lambda for global inline expansion.
+;;;
+;;; Unless a INLINE function, we temporarily clobber the inline
+;;; expansion. This prevents recursive inline expansion of
+;;; opportunistic pseudo-inlines.
+(defun ir1-convert-inline-expansion (var inlinep)
+  (declare (type defined-fun var))
+  (let ((var-expansion (defined-fun-inline-expansion var)))
+    (unless (eq inlinep 'inline)
+      (setf (defined-fun-inline-expansion var) nil))
+    (let* ((name (leaf-source-name var))
+           (fun (ir1-convert-inline-lambda var-expansion
+                                           :source-name name))
+           (info (info :function :info name)))
+      (setf (functional-inlinep fun) inlinep)
+      (assert-new-definition var fun)
+      (setf (defined-fun-inline-expansion var) var-expansion)
+      ;;
+      ;; If definitely not an interpreter stub, then substitute for any
+      ;; old references.
+      (unless (or (eq (defined-fun-inlinep var) 'notinline)
+                  (not (block-compile *compilation*))
+                  (and info
+                       (or (fun-info-transforms info)
+                           (fun-info-templates info)
+                           (fun-info-ir2-convert info))))
+        (substitute-leaf fun var)
+        ;; If in a simple environment, then we can allow backward
+        ;; references to this function from following top-level
+        ;; forms.
+        (when (simple-lexical-environment-p *lexenv*)
+          (setf (defined-fun-functional var) fun)))
+      fun)))
+
 
 ;;; Entry point utilities
 
@@ -1442,42 +1416,32 @@ is potentially harmful to any already-compiled callers using (SAFETY 0)."
 ;;; * a possibly empty list of dynamic extent arguments.
 ;;; The inline lambda will be NIL for a structure accessor, predicate, or copier
 ;;; since those can always be reconstructed from a defstruct description.
-(defun %compiler-defun (name compile-toplevel inline-lambda extra-info)
-  (let ((defined-fun nil)) ; will be set below if we're in the compiler
-    (cond (compile-toplevel
+(defun %compiler-defun (name compile-toplevel inline-lambda extra-info &optional specialized-xep)
+  (cond (compile-toplevel
+         (let ((defined-fun nil))
            (with-single-package-locked-error
                (:symbol name "defining ~S as a function")
-             (setf defined-fun
-                   ;; Try to pass the lambda-list to GET-DEFINED-FUN if we can.
-                   (if (atom inline-lambda)
-                       (get-defined-fun name)
-                       (get-defined-fun
-                        name (ecase (car inline-lambda)
-                               (lambda-with-lexenv (third inline-lambda))
-                               (lambda (second inline-lambda)))))))
+             (setf defined-fun (get-defined-fun name)))
            (when (boundp '*lexenv*)
              (aver (producing-fasl-file))
-             (if (member name (fun-names-in-this-file *compilation*) :test #'equal)
-                 (warn 'duplicate-definition :name name)
-                 (push name (fun-names-in-this-file *compilation*))))
+             (let ((names (fun-names-in-this-file *compilation*)))
+               (if (hashset-find names name)
+                   (warn 'duplicate-definition :name name)
+                   (hashset-insert names name))))
            ;; I don't know why this is guarded by (WHEN compile-toplevel),
            ;; because regular old %DEFUN is going to call this anyway.
-           (%set-inline-expansion name defined-fun inline-lambda extra-info))
-          ((boundp 'sb-fasl::*current-fasl-group*)
-           (if (member name (sb-fasl::fasl-group-fun-names sb-fasl::*current-fasl-group*) :test #'equal)
+           (%set-inline-expansion name defined-fun inline-lambda extra-info)))
+        ((boundp 'sb-fasl::*current-fasl-group*)
+         (let ((names (sb-fasl::fasl-group-fun-names sb-fasl::*current-fasl-group*)))
+           (if (hashset-find names name)
                (warn 'duplicate-definition :name name)
-               (push name (sb-fasl::fasl-group-fun-names sb-fasl::*current-fasl-group*)))))
+               (hashset-insert names name)))))
 
-    (become-defined-fun-name name)
-
-    ;;
-    ;; If there is a type from a previous definition, blast it, since it is
-    ;; obsolete.
-    (when (and defined-fun (neq :declared (leaf-where-from defined-fun)))
-      (setf (leaf-type defined-fun)
-            ;; FIXME: If this is a block compilation thing, shouldn't
-            ;; we be setting the type to the full derived type for the
-            ;; definition, instead of this most general function type?
-            (specifier-type 'function))))
+  (become-defined-fun-name name)
+  (when specialized-xep
+    (setf (info :function :specialized-xep name) specialized-xep)
+    (let ((xep-name (list* 'sb-impl::specialized-xep name specialized-xep)))
+      (setf (info :function :type xep-name) (specifier-type `(function ,@specialized-xep))
+            (info :function :where-from xep-name) :declared)))
 
   (values))
